#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import time
import cv2

from datetime import datetime

import numpy as np

WHITE	= (255, 255, 255)					# BGR形式
BLACK	= (  0,   0,   0)					# BGR形式(黒)

KERNEL4 = np.array([[ 0,-1, 0],[-1, 5,-1],[ 0,-1, 0]])		# 4近傍 ｼｬｰﾌﾟ化
KERNEL8 = np.array([[-1,-1,-1],[-1, 9,-1],[-1,-1,-1]])		# 8近傍 ｼｬｰﾌﾟ化
k=2.0														# 通常は、k=1。k=9で8近傍 ｼｬｰﾌﾟ化。k=2で少しだけｼｬｰﾌﾟ化気味にする。
k1 = -k/9.0
k2 = 1+(8.0*k)/9.0
UNSHARP = np.array([[k1,k1,k1],[k1,k2,k1],[k1,k1,k1]])		# ｱﾝｼｬｰﾌﾟ(鮮鋭化)

################################################
class ZoomFilter :
	#############################################################################################
	def __init__(self):
		'''
			ｺﾝｽﾄﾗｸﾀ
		'''

		# ｸﾘｯｸされた位置の辞書 {'width':'画像幅','height':'画像高さ','stX':'開始X','stY':'開始Y','edX':'終了X','edY':'終了Y',}
		self.areaDic = {}
		self.oldimg  = None					# 画像合成の初期値と初回ﾌﾗｸﾞを兼用
		self.hwTrim  = None					# 画像切り出し領域
		self.clahe   = cv2.createCLAHE(clipLimit=2.0, tileGridSize=(8,8))	# ﾉｲｽﾞ抑制(OpenCVのﾃﾞﾌｫﾙﾄでは40)

	#############################################################################################
	# bottle からﾙｰﾃｨﾝｸﾞされるﾒｿｯﾄﾞ
	def action(self, actDic):
		'''
			bottleｻｰﾊﾞｰから /control ﾘｸｴｽﾄ時の処理

			:param actDic : {'width':'画像幅','height':'画像高さ',
							'stX':'開始X','stY':'開始Y','edX':'終了X','edY':'終了Y',} 形式の辞書
		'''

		# 開始-終了位置が逆転している場合は、初期化する
		stX = int(actDic.get('stX','0'))
		stY = int(actDic.get('stY','0'))
		edX = int(actDic.get('edX','0'))
		edY = int(actDic.get('edY','0'))

		if stX < edX and stY < edY :
			self.areaDic = actDic
		else :
			self.areaDic = {}

		self.hwTrim = None					# 初期化
		self.compFilter = CompositeFilter()

	#############################################################################################
	# filterｸﾗｽとしては必須ﾒｿｯﾄﾞ
	def filter(self, image):
		'''
			openCvの image を入力して、加工処理を行い、image を返します。

			:param  image : 入力ｲﾒｰｼﾞ
			:return	:	加工処理が行われたｲﾒｰｼﾞ
		'''
		try :
			image = self._puttime(image)				# 時刻表示

			if len(self.areaDic) > 0 :					# 縮小処理 有り
				if self.hwTrim is None :				# 縮小処理が初めて実行された時
					h,w,c = image.shape[:3]				# height, width, channels
					fx = self.areaDic['width']  / w		# 原画とhtml表示ｲﾒｰｼﾞとの倍率
					fy = self.areaDic['height'] / h		#

					h1 = int(self.areaDic['stY'] / fy )	# 縦方向（上）
					h2 = int(self.areaDic['edY'] / fy )	# 縦方向（下）
					w1 = int(self.areaDic['stX'] / fx )	# 横方向（左）
					w2 = int(self.areaDic['edX'] / fx )	# 横方向（右）

					self.hwTrim = (h1,h2,w1,w2)
					print( '    filter={}'.format( self.hwTrim ))

				# 切り出し[縦方向（上）：縦方向（下）, 横方向（左）：横方向（右）
				h1,h2,w1,w2 = self.hwTrim
				img = image[h1:h2,w1:w2]				# ｲﾒｰｼﾞの切り出し
				img = self._histogram(img)				# 適用的ﾋｽﾄｸﾞﾗﾑ平坦化
				image = self.compFilter.filter(img)		# 累積画像と新規画像を､合成する

			return image

		except Exception as ex:
			print( 'filter 実行中に例外が発生しました｡' )
			raise ex											# 例外を投げる｡

	#############################################################################################
	# 日時文字列書き込み処理
	def _puttime(self, img):
		now = datetime.today()
		ymd = now.strftime( '%Y/%m/%d %H:%M:%S' )			# 年/月/日 時:分:秒

		# 本来は文字の表示領域のﾋﾟｸｾﾙを求めて計算すべきだが…
		h,w,c = img.shape[:3]								# height, width, channels
		fsize = w/3/180										# 180pxが､ｻｲｽﾞ 1 基準で､1/3領域に表示する(決め打ち)
		hpos  = int(20*fsize)								# 表示位置は､ｻｲｽﾞ 1 基準で､20px で計算する(決め打ち)

		# 時間表示	画像,文字,座標(x,y),ﾌｫﾝﾄ                  ,ｻｲｽﾞ ,色    ,太さ,線の種類
		cv2.putText(img ,ymd ,(0,hpos) ,cv2.FONT_HERSHEY_PLAIN,fsize,WHITE ,2   ,cv2.LINE_AA)	# 少し太い文字で背景を書く
		cv2.putText(img ,ymd ,(0,hpos) ,cv2.FONT_HERSHEY_PLAIN,fsize,BLACK ,1   ,cv2.LINE_AA)	# その中に､本来の文字を書けば見やすくなる｡

		return img

	#############################################################################################
	# 適用的ﾋｽﾄｸﾞﾗﾑ平坦化処理
	def _histogram(self, img):
#		img_yuv = cv2.cvtColor(img, cv2.COLOR_BGR2YUV)		# YCbCr 形式に変更
#		img_yuv[:,:,0] = self.clahe.apply(img_yuv[:,:,0])	# Y:輝度 だけﾋｽﾄｸﾞﾗﾑ平坦化(適用的ﾋｽﾄｸﾞﾗﾑ平坦化)
#		img = cv2.cvtColor(img_yuv, cv2.COLOR_YUV2BGR)		# ｶﾗｰに戻す(YUV→BGR時に色のロスが発生する)

		img_hsv = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)		# HSV 形式に変更
		img_hsv[:,:,2] = self.clahe.apply(img_hsv[:,:,2])	# V:明度 だけﾋｽﾄｸﾞﾗﾑ平坦化(適用的ﾋｽﾄｸﾞﾗﾑ平坦化)
		img = cv2.cvtColor(img_hsv, cv2.COLOR_HSV2BGR)		# ｶﾗｰに戻す

		return cv2.filter2D(img, -1, UNSHARP)				# 4/8近傍ｼｬｰﾌﾟ化,ｱﾝｼｬｰﾌﾟ(鮮鋭化)

	#############################################################################################
	# 終了時に呼ばれるﾒｿｯﾄﾞ
	def close(self):
		'''
			内部変数を初期化します。
		'''
		self.areaDic = {}
		self.hwTrim = None					# 初期化
		self.compFilter.clear()				# 初期化

################################################
class CompositeFilter :
	''' 画像を比重を付けて合成します。 '''
	#############################################################################################
	def __init__(self):
		self.oldimg = None					# 画像合成の初期値

	#############################################################################################
	# filterｸﾗｽとしては必須ﾒｿｯﾄﾞ
	def filter(self, image):
		if self.oldimg is not None :
			image = cv2.addWeighted(self.oldimg,0.9,image,0.1,0)	# 累積画像と新規画像を､9:1 で合成

		self.oldimg = image
		return image

	#############################################################################################
	# 初期化ﾒｿｯﾄﾞ
	def clear(self):
		self.oldimg = None					# 初期化

################################################
# ogServer.py で、bottle/flask に組み込むためのﾌｯｸ
def new_obj():
	'''
		filter/module/subapp処理を行うｸﾗｽのｲﾝｽﾀﾝｽを生成します｡

		:return	:	(filterｲﾝｽﾀﾝｽ,moduleｲﾝｽﾀﾝｽ,subappｲﾝｽﾀﾝｽ)の辞書
	'''
	import json									# ajaxﾃﾞｰﾀ
	import bottle								# ｱﾌﾟﾘｹｰｼｮﾝｻｰﾊﾞｰ

	zoom = ZoomFilter()
	subapp = bottle.Bottle()

	###### ajax 通信 #####
	@subapp.route('/filter' , method='POST')		# bottleは､『method』
	def filterAction():
		# /filter ﾘｸｴｽﾄ時に送られてきた JSON形式を辞書に変換
		actDic = json.load(bottle.request.body)
		zoom.action(actDic)
		return ''								# 非同期通信なので､処理完了時点で値は拾えない

	return {'filter':zoom,'subapp':subapp }

################################################
# mjpg-streamer の input_opencv.so filterの 互換処理
#def init_filter():
#	'''
#		filter処理を行うﾒｿｯﾄﾞを返します。
#
#		:return	:	filterﾒｿｯﾄﾞ
#	'''
#	zoom = ZoomFilter()
#	return zoom.filter

################################################
class MouseClick :
	def __init__(self,fname,width,height,action):
		self.stX = 0						# ｸﾘｯｸ時のX値
		self.stY = 0						# ｸﾘｯｸ時のY値
		self.action = action				# ｸﾘｯｸ時のaction
		self.width  = width					# 画像の幅
		self.height = height				# 画像の高さ

		cv2.namedWindow(fname)						# Window名を指定しないと、ｲﾍﾞﾝﾄが発生しない
		cv2.setMouseCallback(fname,self.mouseClick)	# ﾏｳｽｲﾍﾞﾝﾄのCallback関数登録

	def mouseClick(self,event,x,y,flags,param):
		if event == cv2.EVENT_LBUTTONDOWN:			# 左ﾎﾞﾀﾝを押す
			print('DOWN x={},y={}'.format(x,y))
			self.stX = x							# X座標(開始点)
			self.stY = y							# Y座標(開始点)
		elif event == cv2.EVENT_LBUTTONUP:			# 左ﾎﾞﾀﾝを離す
			print('UP   x={},y={}'.format(x,y))
			actDic = {	'width' :self.width	,
						'height':self.height,
						'stX':self.stX ,
						'stY':self.stY ,
						'edX':x ,					# X座標(終了点)
						'edY':y						# Y座標(終了点)
					}
			self.action(actDic)

################################################
# 動作確認用のテストプログラムです。
def main(argDic):
	filter = ZoomFilter()						# ｵﾌﾞｼﾞｪｸﾄ生成

	fname = argDic.get('fname')					# ﾌﾟﾛｸﾞﾗﾑ名=Windows画面名
	devno = int(argDic.get('devno'))			# ﾃﾞﾊﾞｲｽNoの設定

	cap = cv2.VideoCapture(devno)
	width	= int( cap.get(cv2.CAP_PROP_FRAME_WIDTH)  )	# 画像の幅
	height	= int( cap.get(cv2.CAP_PROP_FRAME_HEIGHT) )	# 画像の高さ

	MouseClick(fname,width=width,height=height,action=filter.action)

	try:
		while True:
			_,img = cap.read()
			img = filter.filter(img)
			cv2.imshow(fname,img)

			key = cv2.waitKey(10)				# ミリ秒単位で表される遅延時間
			if key == 27:						# ESCを押した時
				break

			time.sleep( 0.1 )					# CPU処理の軽減化

	except KeyboardInterrupt  : 				# Ctl+Cが押されたらループを終了
		print( "\nCtl+C Stop" )
	except Exception as ex:
		print( ex )								# 例外処理の内容をコンソールに表示
		import traceback
		traceback.print_exc()					# Exception のトレース
	finally :
		if cap is not None : cap.release()
		cv2.destroyAllWindows()
		print( "終了" )

################################################
# main関数を呼び出します
################################################
if __name__=='__main__':
	print( 'python zoom_filter.py [devno(=0)]' )

	argDic = {'fname':None , 'devno':'0'}

	import sys								# 引数の読み込み用
	args = sys.argv
	argDic['fname'] = args[0]
	if len(args) > 1 : argDic['devno'] = args[1]

	main(argDic)
